/* A Programmer's Macro Pad */
/* CC BY-NC-SA Jeroen Brinkman */
/* Release 1 - 2/4/2025 */
/* Improved version, emergency stop added*/

/*--------------------------------------------------------------*/
/* Use the Adrduino Leonardo for programming, not the Pro!      */
/* Unplug the Arduino pro and replug when compiling has started */
/* Or attach a reset button and press the reset button while    */
/* compiling and release this button when                       */
/* "Waiting for upload port..." is shown.                       */
/* Connect A3 to ground, to cancel all keyboard presses.        */
/* Macro Pad is ready for use when blue light is off.           */
/*--------------------------------------------------------------*/

#include <Keypad.h>      // https://playground.arduino.cc/Code/Keypad/ The input
#include <Keyboard.h>    // https://docs.arduino.cc/language-reference/en/functions/usb/Keyboard/ The output

/* standard definitions */
#define ROWS         4   // Number of rows
#define COLS         4   // Number of columns
#define BAUD      9600   // Baud rate Serial port
#define WAIT        10   // Number of milliseconds to wait after a character is send
#define SIZE       800   // Size array of strings; be sure this is long enough! If too short, the power LED will blink
#define ON           1   // Power LED on
#define OFF          0   // Power LED off

/* special characters used in the strings, see fill procedure*/
#define LEFT       '<'   // Left arrow key symbol
#define UP         '^'   // Upper arrow key symbol
#define RIGHT      '>'   // Right arrow key symbol
#define DELETE     '#'   // Delete key symbol
#define ENTER      '~'   // Enter key symbol

/* input definitions */
#define IO_POWER    10   // Power LED for the Arduino
#define IO_ENABLE   A3   // If put to ground, the Arduino will not type (emergency brake)

// Define the Keymap, adjust the positions of '0'-'F' if needed
char keys[ROWS][COLS] = {{'0','1','2','3'}, // Change if wirering requires
                         {'4','5','6','7'},
                         {'8','9','A','B'},
                         {'C','D','E','F'}};

byte rowPins[ROWS] = {2, 3, 4, 5}; //connect to the row pinouts of the keypad, make grid logical sequence from 1-9 independent from pins used
byte colPins[COLS] = {6, 7, 8, 9}; //connect to the column pinouts of the keypad, there are none as it's solved with separate pins per button

char text[SIZE];         // The string array
int index[16][2];        // The array containing start and end position of each
char key;                // The character pressed
int press,               // The actual key press
lastpress,               // The last key press position
      i,j,               // indices
      pos;               // The character position
bool error;              // The number of characters exceeds the SIZE

Keypad keypad = Keypad( makeKeymap(keys), rowPins, colPins, ROWS, COLS );

int convert(char c) {
  int v;
  switch (c) {
    case 'A': 
      v = 10;
      break;
    case 'B': 
      v = 11;
      break;
    case 'C': 
      v = 12;
      break;
    case 'D': 
      v = 13;
      break;
    case 'E': 
      v = 14;
      break;
    default:
      v = c - '0';
  }
  return v;
} // convert

void type(int k) { // Special characters can be coded in the string and pressed

  if (digitalRead(IO_ENABLE)) {
    switch (text[k]) {
      case RIGHT: 
        Keyboard.press(KEY_RIGHT_ARROW);
        break;
      case UP: 
        Keyboard.press(KEY_UP_ARROW);
        break;
      case LEFT: 
        Keyboard.press(KEY_LEFT_ARROW);
        break;
      case ENTER: 
        Keyboard.press(KEY_RETURN);
        break;
      case DELETE: 
        Keyboard.press(KEY_DELETE);
        break;
      default:
        Keyboard.print(text[k]);
    } // if
  } // switch
  Keyboard.releaseAll();
  delay(WAIT); // wait a while to process the send character
} // type

void set(String store) {

  if (press != lastpress) {
    index[press][0] = pos; // start position of this string in text
    lastpress = press;
  }
  for(i = 0; i < store.length(); i++) { 
    text[pos] = store.charAt(i);
    pos += 1;
    if (pos == SIZE) { // generate error when allocated text SIZE is not enough
      error = true;
      return;
    }
  }
  index[press][1] = pos; // end position + 1 of this string in text
} // set
  
void fill() {
  // Define your own strings as required
  // <left arrow, ^top arrow, >right arrow, #delete, ~enter
  press = 0; // Log
  set("Login~"); 
  press = 1; // Function
  set("int () { // ~"); 
  set("   ;~"); 
  set("  return 0; ~"); 
  set("}; // func~^^^<<<<<<<<"); 
  press = 2; // Procedure
  set("void () { // ~"); 
  set("   ;~"); 
  set("}; // ~<^^"); 
  press = 3; // Array
  set("int array[][] = {{},{}};  // ~^>>>>>>>>>>"); 
  press = 4; // if
  set("  if ( == 0) { // when ~"); 
  set("    ;~"); 
  set("  }; // if~^^^>>>>>>"); 
  press = 5; // else
  set("  } else { // ~"); 
  set("    ;~^>>>>"); 
  press = 6; // = if
  set("  = ( == ) ? : ;~^>>~"); 
  press = 7; // switch
  set("  switch () {~"); 
  set("    case 1: // ~"); 
  set("      ;~"); 
  set("      break; ~"); 
  set("    case 2: // ~"); 
  set("      ;~"); 
  set("      break; ~"); 
  set("    case 3: // ~"); 
  set("      ;~"); 
  set("      break; ~"); 
  set("    default:~"); 
  set("      ;~"); 
  set("  }; // switch~^^^^^^^^^^^^<<<<"); 
  press = 8; // for
  set("  for (i = ; i < ; i++) { // ~"); 
  set("    ;~"); 
  set("  }; // for~<^^"); 
  press = 9; // while
  set("  while ( == ) { // ~"); 
  set("    ;~"); 
  set("  }; // while~<^^<<<<"); 
  press = 10; // Digital Read
  set("    digitalRead(, ); // ~<<<<<<<<<"); 
  press = 11; // Digital Write
  set("    digitalWrite(, ); // ~<<<<<<<<<"); 
  press = 12; // Serial print
  set("    Serial.println("""");~"); 
  set("    Serial.print("""");~<<<<"); 
  press = 13; // Tab
  set("  <<^"); 
  press = 14; // Del
  set("##^"); 
  press = 15; // Three lines
  set("    ; // ~"); 
  set("    ; ~"); 
  set("    ;~^^<<<<<<"); 
}; // fill

void setup() {
  /* initialise the I/O ports */
  pinMode(IO_POWER, OUTPUT);
  digitalWrite(IO_POWER, ON); // power LED on
  pinMode(IO_ENABLE, INPUT_PULLUP);

  /* clear the matrix */
  for (i = 0; i < 16; i++) {
    index[i][0] = 0; 
    index[i][1] = 0; 
  } // for

  /* fill the matrix */
  press = 0; // the actual programmable key
  lastpress = 0; // the last programmable key
  pos = 0; // the actual line
  error = false;
  fill();

  /* initialise */
  Serial.begin(BAUD);
  if (digitalRead(IO_ENABLE)) Keyboard.begin();
  delay(20000); // wait a while in order to enable reporgramming of the Arduino
  digitalWrite(IO_POWER, OFF); // Power LED off -> keypad ready
}

void loop() {
  if (error) digitalWrite(IO_POWER, (millis() / 1000) % 2); // 1s error flashing for the power LED

  key = keypad.getKey(); // read the keypress 
  press = convert(key); // convert character to integer

  if ((key) && digitalRead(IO_ENABLE)){ // key pressed
    digitalWrite(IO_POWER, ON); // Power LED on - > keypad transmission in progress
    Keyboard.begin();
    for(i = index[press][0]; i < index[press][1]; i++) type(i); // type the macro text
    Keyboard.releaseAll();
    digitalWrite(IO_POWER, OFF); // Power LED off -> keypad ready
    Keyboard.end();
  } // if
  if (!digitalRead(IO_ENABLE)) Keyboard.end(); // Stop keyboard functions if not enabled
} // loop
